@@ -11,12 +11,6 @@
 #include <stdarg.h>
 #include <unistd.h>
 #include "config.h"
-#ifdef HAVE_GETOPT_LONG
-#include <getopt.h>
-#endif
-#ifdef HAVE_REGEXEC
-#include <regex.h>
-#endif
 #include <string.h>
 #include <ctype.h>
 #include <sys/types.h>
@@ -25,6 +19,15 @@
 #include <netdb.h>
 #include <errno.h>
 #include <signal.h>
+#ifdef HAVE_GETOPT_LONG
+#include <getopt.h>
+#endif
+#ifdef HAVE_REGEXEC
+#include <regex.h>
+#endif
+#ifdef HAVE_LIBIDN
+#include <idna.h>
+#endif
 
 /* Application-specific */
 #include "data.h"
@@ -78,7 +81,7 @@ int main(int argc, char *argv[])
 	}
 	if (strchr(ripeflagsp, ch)) {
 	    for (p = fstring; *p; p++);
-	    sprintf(p--, "-%c %s ", ch, optarg);
+	    snprintf(p--, sizeof(fstring), "-%c %s ", ch, optarg);
 	    if (ch == 't' || ch == 'v' || ch == 'q')
 		nopar = 1;
 	    continue;
@@ -161,9 +164,11 @@ int main(int argc, char *argv[])
 	char *tmp;
 
 	tmp = normalize_domain(qstring);
-	server = whichwhois(tmp);
-	free(tmp);
+	free(qstring);
+	qstring = tmp;
+	server = whichwhois(qstring);
 
+retry:
 	switch (server[0]) {
 	    case 0:
 		if (!(server = getenv("WHOIS_SERVER")))
@@ -219,6 +224,21 @@ int main(int argc, char *argv[])
 	    case 6:
 		puts(_("Unknown AS number or IP network. Please upgrade this program."));
 		exit(0);
+	    case 0x0A:
+		{
+		char *tmp6 = convert_6to4(qstring);
+		free(qstring);
+		qstring = tmp6;
+		printf(_("\nQuerying for the IPv4 endpoint %s of a 6to4 IPv6 address.\n\n"), qstring);
+		server = whichwhois(qstring);
+		/*
+		 * This code sucks enough that I can afford to use goto...
+		 * Some day whichwhois() and queryformat() will be merged
+		 * and will return a struct with status code, server name
+		 * and query string.
+		 */
+		goto retry;
+		}
 	    default:
 		if (verb)
 		    printf(_("Using server %s.\n"), server);
@@ -330,6 +350,8 @@ const char *whichwhois(const char *s)
 		if (v6net == ip6_assign[i].net)
 		    return ip6_assign[i].serv;
 	    return "\x06";			/* unknown allocation */
+	} else if (strncmp(s, "2002:", 5) == 0) {
+	    return "\x0A";
 	} else if (strncasecmp(s, "3ffe:", 5) == 0)
 	    return "whois.6bone.net";
 	/* RPSL hierarchical object like AS8627:fltr-TRANSIT-OUT */
@@ -402,9 +424,9 @@ char *queryformat(const char *server, const char *flags, const char *query)
     char *buf;
     int i, isripe = 0;
 
-    /* +10 for CORE; +2 for \r\n; +1 for NULL */
+    /* +2 for \r\n; +1 for NULL */
     buf = malloc(strlen(flags) + strlen(query) + strlen(client_tag) + 4
-	    + 10 + 2 + 1);
+	    + 2 + 1);
     *buf = '\0';
     for (i = 0; ripe_servers[i]; i++)
 	if (strcmp(server, ripe_servers[i]) == 0) {
@@ -433,8 +455,6 @@ char *queryformat(const char *server, const char *flags, const char *query)
 	    strcmp(server, "whois.nic.ad.jp") == 0) &&
 	    strncasecmp(query, "AS", 2) == 0 && isasciidigit(query[2]))
 	sprintf(buf, "AS %s", query + 2);	/* fix query for DDN */
-    else if (!isripe && strcmp(server, "whois.corenic.net") == 0)
-	sprintf(buf, "--machine %s", query);	/* machine readable output */
     else if (!isripe && strcmp(server, "whois.nic.ad.jp") == 0) {
 	char *lang = getenv("LANG");	/* not a perfect check, but... */
 	if (!lang || (strncmp(lang, "ja", 2) != 0))
@@ -520,8 +540,6 @@ const char *query_crsnic(const int sock, const char *query)
     if (write(sock, temp, strlen(temp)) < 0)
 	err_sys("write");
     while (fgets(buf, sizeof(buf), fi)) {
-	if (strncmp(buf, "No match for \"", 14) == 0)	/* ugly */
-	    fputs(buf, stdout);
 	/* If there are multiple matches only the server of the first record
 	   is queried */
 	if (state == 0 && strncmp(buf, "   Domain Name:", 15) == 0)
@@ -537,8 +555,10 @@ const char *query_crsnic(const int sock, const char *query)
 	    *q = '\0';
 	    state = 2;
 	}
-	if (verb)
-	    fputs(buf, stdout);
+	/* the output must not be hidden or no data will be shown for
+	   host records and not-existing domains */
+	/* XXX feel free to send a patch to hide the long disclaimer */
+	fputs(buf, stdout);
     }
     if (ferror(fi))
 	err_sys("fgets");
@@ -590,7 +610,7 @@ const char *query_pir(const int sock, const char *query)
 
 int openconn(const char *server, const char *port)
 {
-    int fd;
+    int fd = -1;
 #ifdef HAVE_GETADDRINFO
     int err;
     struct addrinfo hints, *res, *ai;
@@ -674,9 +694,29 @@ char *normalize_domain(const char *dom)
     for (p = ret; *p; p++); p--;	/* move to the last char */
     for (; *p == '.' || p == ret; p--)	/* eat trailing dots */
 	*p = '\0';
+
+#ifdef HAVE_LIBIDN
+    if (idna_to_ascii_from_locale(ret, &p, 0, 0) != IDNA_SUCCESS)
+	return ret;
+
+    free(ret);
+    ret = p;
+#endif
     return ret;
 }
 
+char *convert_6to4(const char *s)
+{
+    char *new = malloc(sizeof("255.255.255.255"));
+    unsigned int a, b;
+
+    if (sscanf(s, "2002:%x:%x:", &a, &b) != 2)
+	return (char *) "0.0.0.0";
+
+    sprintf(new, "%d.%d.%d.%d", a >> 8, a & 0xff, b >> 8, b & 0xff);
+    return new;
+}
+
 unsigned long myinet_aton(const char *s)
 {
     int a, b, c, d;
